import asyncio
import logging
import coloredlogs

LOG_FORMAT = "[%(asctime)s][%(levelname)s][%(name)s] %(message)s"
FIELD_STYLE = {
    "asctime": {"color": "green"},
    "levelname": {"color": "blue", "bold": True},
    "name": {"color": "magenta"},
    "message": {"color": 144, "bright": False},
}


class AsyncLogger(object):
    def __init__(self, name: str = "async-logger", level: str = "INFO", queue_size: int = 1000) -> None:
        self._sync_logger = logging.getLogger(f"{name}-sync")
        coloredlogs.install(level=level, logger=self._sync_logger, fmt=LOG_FORMAT, field_styles=FIELD_STYLE)
        self.logger = logging.getLogger(name)
        self.logger.setLevel(level)
        self.queue: asyncio.Queue[logging.LogRecord | None] = asyncio.Queue(maxsize=queue_size)
        self.logger.handlers.clear()
        self.logger.addHandler(_AsyncQueueHandler(self.queue))

        self._task: asyncio.Task | None = None

    async def start(self) -> None:
        if not self._task or self._task.done():
            self._task = asyncio.create_task(self._log_worker())

    async def stop(self) -> None:
        if self._task:
            await self.queue.put(None)
            await self._task
            self._task = None

    async def _log_worker(self) -> None:
        while True:
            record = await self.queue.get()
            if record is None:
                break
            self._sync_logger.handle(record)

    async def debug(self, *args, **kwargs) -> None:
        self.logger.debug(*args, **kwargs)

    async def info(self, *args, **kwargs) -> None:
        self.logger.info(*args, **kwargs)

    async def warning(self, *args, **kwargs) -> None:
        self.logger.warning(*args, **kwargs)

    async def error(self, *args, **kwargs) -> None:
        self.logger.error(*args, **kwargs)

    async def critical(self, *args, **kwargs) -> None:
        self.logger.critical(*args, **kwargs)

    async def exception(self, *args, **kwargs) -> None:
        self.logger.exception(*args, **kwargs)


class _AsyncQueueHandler(logging.Handler):
    def __init__(self, queue: asyncio.Queue) -> None:
        super().__init__()
        self.queue = queue

    def emit(self, record: logging.LogRecord) -> None:
        try:
            self.queue.put_nowait(record)
        except asyncio.QueueFull:
            pass


main_logger = AsyncLogger(name="ChunithmMusicScanner", level="DEBUG")
scanner_logger = AsyncLogger("MusicDataScanner", level="DEBUG")



from sqlalchemy.orm import DeclarativeBase
from sqlalchemy import (
    Column,
    String,
    Integer,
    Float,
    Date,
    CheckConstraint,
    Numeric,
)


class ChunithmMusicDBBase(DeclarativeBase):
    __abstract__ = True


class ChunithmChartData(ChunithmMusicDBBase):
    __tablename__ = "chart_data"
    music_id = Column(Integer, primary_key=True)
    difficulty = Column(Integer, primary_key=True)
    creator = Column(String(50), nullable=True)
    bpm = Column(Float, nullable=True)
    tap_count = Column(Integer, nullable=True)
    hold_count = Column(Integer, nullable=True)
    slide_count = Column(Integer, nullable=True)
    air_count = Column(Integer, nullable=True)
    flick_count = Column(Integer, nullable=True)
    total_count = Column(Integer, nullable=True)


class ChunithmMusic(ChunithmMusicDBBase):
    __tablename__ = "music"
    music_id = Column(Integer, primary_key=True)
    title = Column(String(255), nullable=False)
    artist = Column(String(255), nullable=False)
    category = Column(String(50), nullable=True)
    version = Column(String(10), nullable=True)
    release_date = Column(Date, nullable=True)
    is_deleted = Column(Integer, CheckConstraint("is_deleted IN (0, 1)"), default=0)
    deleted_version = Column(String(10), nullable=True)


class ChunithmMusicDifficulty(ChunithmMusicDBBase):
    __tablename__ = "music_difficulties"
    music_id = Column(Integer, primary_key=True)
    version = Column(String(10), primary_key=True)
    diff0_const = Column(Numeric(3, 1))
    diff1_const = Column(Numeric(3, 1))
    diff2_const = Column(Numeric(3, 1))
    diff3_const = Column(Numeric(3, 1))
    diff4_const = Column(Numeric(3, 1))



import sys
import json
import asyncio
import argparse
import traceback

from aiopath import AsyncPath

from .scanner import scan_music
from .sa.engine import DatabaseEngine
from .logger import main_logger as logger, scanner_logger
from .sa.tables import ChunithmMusicDBBase, ChunithmMusic, ChunithmMusicDifficulty, ChunithmChartData


engine: DatabaseEngine | None = None
a000_path: AsyncPath | None = None
options_dir: AsyncPath | None = None
deleted_option_path: AsyncPath | None = None


async def init(config_path: AsyncPath) -> None:
    if not await config_path.is_file():
        await logger.error(f"Config file not found: {config_path}")
        sys.exit(1)

    global engine, a000_path, options_dir, deleted_option_path
    content = await config_path.read_text(encoding="utf-8")
    data = json.loads(content)
    try:
        a000_path, options_dir, deleted_option_path = (
            data["a000_path"],
            data["options_dir"],
            data["deleted_option_path"],
        )
        a000_path, options_dir, deleted_option_path = (
            AsyncPath(a000_path) if a000_path else None,
            AsyncPath(data["options_dir"]) if options_dir else None,
            AsyncPath(data["deleted_option_path"]) if deleted_option_path else None,
        )
        await logger.info("Loaded A000, options directory and deleted option configuration.")
    except Exception as e:
        traceback.print_exc()
        await logger.error("Failed to load A000, options, deleted option directory configuration.")
        sys.exit(1)
    try:
        await logger.info("Initializing database engine...")
        db_host, db_port, db_user, db_pass, db_name = (
            data["host"],
            data["port"],
            data["user"],
            data["password"],
            data["database"],
        )
        engine = DatabaseEngine(
            url_scheme="mysql+asyncmy://{user}:{password}@{host}:{port}/{database}".format(
                user=db_user, password=db_pass, host=db_host, port=db_port, database=db_name
            ),
            table_base=ChunithmMusicDBBase,
        )
        await engine.init_engine()
        await logger.info("Initialized database engine.")
    except Exception as e:
        traceback.print_exc()
        await logger.error("Failed to initialize database engine.")
        sys.exit(1)


async def update_database() -> None:
    music_infos, music_diffs, chart_data = await scan_music(a000_path, options_dir, deleted_option_path)
    music_infos = [ChunithmMusic(**(m.model_dump(exclude_none=True))) for m in music_infos]
    music_diffs = [ChunithmMusicDifficulty(**(m.model_dump(exclude_none=True))) for m in music_diffs]
    chart_data = [ChunithmChartData(**(m.model_dump(exclude_none=True))) for m in chart_data]
    await logger.info("Upserting music data to database...")
    await engine.upsert(music_infos)
    await engine.upsert(music_diffs)
    await engine.upsert(chart_data)
    await logger.info("Upserted music data to database.")


async def async_main():
    await logger.start()
    await scanner_logger.start()
    parser = argparse.ArgumentParser(description="Chunithm Music Scanner")
    parser.add_argument(
        "--config",
        type=AsyncPath,
        default=AsyncPath("configs.json"),
        help="Path to configs.json (default: ./configs.json)",
    )

    args = parser.parse_args()
    await init(args.config)
    await update_database()
    await logger.info("Terminating Chunithm Music Scanner...")
    await scanner_logger.stop()
    await logger.stop()


def main():
    asyncio.run(async_main())


if __name__ == "__main__":
    main()
    sys.exit(0)



from datetime import date
from pydantic import BaseModel, ConfigDict


class MusicInfoSchema(BaseModel):
    music_id: int
    title: str
    artist: str
    category: str | None = None
    version: str | None = None
    release_date: date | None = None
    is_deleted: bool | None = None
    deleted_version: str | None = None

    model_config = ConfigDict(from_attributes=True)


class MusicDifficultySchema(BaseModel):
    music_id: int
    version: str
    diff0_const: float | None = None
    diff1_const: float | None = None
    diff2_const: float | None = None
    diff3_const: float | None = None
    diff4_const: float | None = None

    model_config = ConfigDict(from_attributes=True)


class ChartDataSchema(BaseModel):
    music_id: int
    difficulty: int
    creator: str | None = None
    bpm: float | None = None
    tap_count: int | None = None
    hold_count: int | None = None
    slide_count: int | None = None
    air_count: int | None = None
    flick_count: int | None = None
    total_count: int | None = None

    model_config = ConfigDict(from_attributes=True)



from sqlalchemy import select
from sqlalchemy.inspection import inspect
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import Union, List, Type, TypeVar, Sequence
from sqlalchemy.dialects.mysql import insert as mysql_insert
from sqlalchemy.orm import InstrumentedAttribute, DeclarativeBase
from sqlalchemy.ext.asyncio import create_async_engine, async_sessionmaker, AsyncSession


T = TypeVar("T")


class DatabaseEngine:
    def __init__(self, url_scheme, table_base: Type[DeclarativeBase]) -> None:
        self._engine = create_async_engine(url_scheme, echo=False, future=True)
        self._session_maker = async_sessionmaker(self._engine, expire_on_commit=False)
        self._table_base = table_base

    async def init_engine(self):
        async with self._engine.begin() as conn:
            await conn.run_sync(self._table_base.metadata.create_all)

    @asynccontextmanager
    async def session(self) -> AsyncGenerator[AsyncSession | None]:
        async with self._session_maker() as _session:
            yield _session

    async def shutdown_engine(self) -> None:
        await self._engine.dispose()

    async def select(
        self,
        target: Union[Type[T], InstrumentedAttribute],
        *conditions,
        one_result: bool = False,
    ) -> Union[T, List[T]] | None:
        async with self.session() as session:
            stmt = select(target)
            if conditions:
                stmt = stmt.where(*conditions)
            result = await session.execute(stmt)
            if one_result:
                return result.scalar_one_or_none()
            return result.scalars().all()

    async def add(self, instances: T | Sequence[T]) -> T | list[T] | None:
        async with self.session() as session:
            if isinstance(instances, list):
                session.add_all(instances)
            else:
                session.add(instances)
            await session.commit()
            if isinstance(instances, list):
                for inst in instances:
                    await session.refresh(inst)
                return instances
            else:
                await session.refresh(instances)
                return instances

    async def upsert(self, instances: T | Sequence[T]) -> T | list[T] | None:
        if not isinstance(instances, list):
            instances = [instances]
            single = True
        else:
            single = False
        if not instances:
            return [] if not single else None
        model = type(instances[0])
        table = model.__table__
        pk_columns = [col.name for col in inspect(model).primary_key]
        if not pk_columns:
            raise ValueError("Model has no primary key defined")
        values_list = []
        update_keys_list = []
        for instance in instances:
            values = {col.name: getattr(instance, col.name) for col in table.columns if hasattr(instance, col.name)}
            values_list.append(values)
            update_keys_list.append([
                k for k, v in values.items()
                if k not in pk_columns and v is not None
            ])
        async with self.session() as session:
            stmt = mysql_insert(table).values(values_list)
            common_update_keys = set.intersection(*map(set, update_keys_list))
            stmt = stmt.on_duplicate_key_update({
                k: stmt.inserted[k] for k in common_update_keys
            })
            await session.execute(stmt)
            await session.commit()
            result_list = []
            for values in values_list:
                filter_kwargs = {k: values[k] for k in pk_columns}
                result = await session.execute(select(model).filter_by(**filter_kwargs))
                result_list.append(result.scalar_one_or_none())
            return result_list[0] if single else result_list



import re
import asyncio
import configparser
from typing import Any
from pathlib import Path

from aiopath import AsyncPath
from datetime import datetime, date
import xml.etree.ElementTree as xmlEt

from .logger import scanner_logger
from .schemas import MusicDifficultySchema, MusicInfoSchema, ChartDataSchema

c2s_mapping = {
    "CREATOR": "creator",
    "BPM_DEF": "bpm",
    "T_JUDGE_TAP": "tap_count",
    "T_JUDGE_HLD": "hold_count",
    "T_JUDGE_SLD": "slide_count",
    "T_JUDGE_AIR": "air_count",
    "T_JUDGE_FLK": "flick_count",
    "T_JUDGE_ALL": "count",
}


def parse_music_xml(
    xml_path: AsyncPath, current_version: str, is_deleted: bool = False
) -> tuple[MusicInfoSchema, MusicDifficultySchema]:
    tree = xmlEt.parse(Path(xml_path))
    root = tree.getroot()

    def find_text(path: str) -> str | None:
        el = root.find(path)
        return el.text if el is not None and el.text else None

    def parse_release_date(raw: str | None) -> date | None:
        if not raw or len(raw) != 8:
            return None
        try:
            return datetime.strptime(raw, "%Y%m%d").date()
        except ValueError:
            return None

    def extract_version_tag(tag_str: str | None) -> str | None:
        if not tag_str:
            return None
        match = re.search(r"(\d+)\.(\d+)", tag_str)
        return match.group(0) if match else None

    def get_fumen_level_by_type_id(type_id: int) -> float | None:
        for fumen in root.findall("./fumens/MusicFumenData"):
            fumen_type_id = fumen.findtext("./type/id")
            if fumen_type_id and int(fumen_type_id) == type_id:
                level = fumen.findtext("level") or "0"
                decimal = fumen.findtext("levelDecimal") or "0"
                return float(f"{level}.{decimal}")
        return None

    music_id = int(find_text("./name/id") or 0)
    title = find_text("./name/str") or ""
    artist = find_text("./artistName/str") or ""
    category = find_text("./genreNames/list/StringID/str")
    release_tag = find_text("./releaseTagName/str")
    release_version = extract_version_tag(release_tag)
    release_date = parse_release_date(find_text("./releaseDate"))

    info = MusicInfoSchema(
        music_id=music_id,
        title=title,
        artist=artist,
        category=category,
        version=release_version,
        release_date=release_date,
        is_deleted=is_deleted,
    )

    diff = MusicDifficultySchema(
        music_id=music_id,
        version=current_version if current_version != "0.00" else release_version,
        diff0_const=get_fumen_level_by_type_id(0),
        diff1_const=get_fumen_level_by_type_id(1),
        diff2_const=get_fumen_level_by_type_id(2),
        diff3_const=get_fumen_level_by_type_id(3),
        diff4_const=get_fumen_level_by_type_id(4),
    )

    return info, diff


async def c2s_analyzer(file_path: AsyncPath) -> ChartDataSchema:
    data: dict[str, Any] = {}
    match = re.match(r"(\d{4})_(\d{2})\.c2s$", file_path.name)
    if not match:
        raise ValueError(f"Invalid c2s filename format: {file_path.name}")
    data["music_id"] = int(match.group(1))
    data["difficulty"] = int(match.group(2))
    async with file_path.open("r", encoding="utf-8") as f:
        async for line in f:
            parts = line.strip().split(maxsplit=1)
            if len(parts) < 2:
                continue
            identifier = parts[0]
            value = parts[1]
            if identifier in c2s_mapping:
                field = c2s_mapping[identifier]
                if identifier == "BPM_DEF":
                    data[field] = float(value.split()[0])
                else:
                    data[field] = value
    await scanner_logger.debug(f"Scanned c2s data: {str(file_path)}")
    return ChartDataSchema(**data)


async def option_ver(option_path: AsyncPath | None = None) -> str | None:
    if not option_path:
        return None
    config_path = option_path / "data.conf"
    if not await config_path.is_file():
        return None
    content = await config_path.read_text(encoding="utf-8")
    parser = configparser.ConfigParser()
    parser.read_string(content)
    try:
        major = parser.getint("Version", "VerMajor")
        minor = parser.getint("Version", "VerMinor")
        if major <= 2 or minor <= 55:
            return f"{major}.{minor}"
        else:
            return None
    except (configparser.Error, ValueError):
        return None


async def iter_all_files(dir_path: AsyncPath):
    async for entry in dir_path.iterdir():
        if await entry.is_dir():
            async for sub in iter_all_files(entry):
                yield sub
        else:
            yield entry


async def scan_music(
    a000_path: AsyncPath | None = None,
    options_dir: AsyncPath | None = None,
    deleted_option_path: AsyncPath | None = None,
) -> tuple[list[MusicInfoSchema], list[MusicDifficultySchema], list[ChartDataSchema]]:
    valid_dirs: dict[AsyncPath:str] = {}
    music_infos: list[MusicInfoSchema] = []
    music_diffs: list[MusicDifficultySchema] = []
    chart_data: list[ChartDataSchema] = []
    await scanner_logger.info("Validating option directories...")
    opt_v = await option_ver(a000_path)
    if opt_v:
        valid_dirs[a000_path] = opt_v
    if options_dir:
        async for option_path in options_dir.iterdir():
            if await option_path.is_dir():
                opt_v = await option_ver(option_path)
                if opt_v:
                    valid_dirs[option_path] = opt_v
    if deleted_option_path:
        valid_dirs[deleted_option_path] = "0.00"
    await scanner_logger.info("Validated option directories.")
    await scanner_logger.info("Scanning music data...")
    xml_tasks = []
    c2s_tasks = []
    xml_file_infos = []
    for valid_dir, opt_v in valid_dirs.items():
        async for file in iter_all_files(valid_dir):
            if file.name == "Music.xml":
                is_deleted = valid_dir == deleted_option_path
                xml_file_infos.append((file, opt_v, is_deleted))
            elif file.suffix == ".c2s":
                c2s_tasks.append(c2s_analyzer(file))
    for file, opt_v, is_deleted in xml_file_infos:
        xml_tasks.append(asyncio.to_thread(parse_music_xml, file, opt_v, is_deleted))
    results = await asyncio.gather(*xml_tasks, return_exceptions=True)
    for i, result in enumerate(results):
        file, _, _ = xml_file_infos[i]
        if isinstance(result, Exception):
            await scanner_logger.error(f"[Music.xml error] {file}: {result}")
        else:
            info, diff = result
            music_infos.append(info)
            music_diffs.append(diff)
            await scanner_logger.debug(f"Scanned music data: {file}")

    c2s_results = await asyncio.gather(*c2s_tasks, return_exceptions=True)
    c2s_files = []
    for valid_dir in valid_dirs:
        async for file in iter_all_files(valid_dir):
            if file.suffix == ".c2s":
                c2s_files.append(file)
    for i, result in enumerate(c2s_results):
        file = c2s_files[i] if i < len(c2s_files) else None
        if isinstance(result, Exception):
            await scanner_logger.error(f"[C2S error] {file}: {result}")
        else:
            chart_data.append(result)

    await scanner_logger.info("Scanned music data.")
    return music_infos, music_diffs, chart_data
